import functools

from pdm.models.candidates import Candidate
from pdm.models.requirements import Requirement
from pdm.project import Project

from pdm_conda.models.requirements import CondaRequirement

_patched = False


def wrap_save_version_specifiers(func):
    @functools.wraps(func)
    def wrapper(
        requirements: dict[str, dict[str, Requirement]],
        resolved: dict[str, Candidate],
        save_strategy: str,
    ) -> None:
        func(requirements, resolved, save_strategy)
        for reqs in requirements.values():
            for name, r in reqs.items():
                if isinstance(r, CondaRequirement):
                    r.version_mapping = resolved[name].req.version_mapping
                    r.is_python_package = resolved[name].req.is_python_package

    return wrapper


def wrap_format_lockfile(func):
    @functools.wraps(func)
    def wrapper(
        project: Project,
        mapping: dict[str, Candidate],
        fetched_dependencies: dict[tuple[str, str | None], list[Requirement]],
    ) -> dict:
        for deps in fetched_dependencies.values():
            for i, dep in enumerate(deps):
                if isinstance(dep, CondaRequirement):
                    setattr(dep, "as_line", functools.partial(dep.as_line, with_build_string=True))
        res = func(project, mapping, fetched_dependencies)
        for deps in fetched_dependencies.values():
            for i, dep in enumerate(deps):
                if isinstance(dep, CondaRequirement):
                    setattr(dep, "as_line", dep.as_line.func)  # type: ignore
        return res

    return wrapper


if not _patched:
    from pdm.cli import actions, utils

    save_version_specifiers = wrap_save_version_specifiers(utils.save_version_specifiers)
    format_lockfile = wrap_format_lockfile(utils.format_lockfile)
    for m in [utils, actions]:
        setattr(m, "save_version_specifiers", save_version_specifiers)
        setattr(m, "format_lockfile", format_lockfile)
    _patched = True



from typing import Collection, cast

from pdm.installers import Synchronizer
from pdm.models.candidates import Candidate

from pdm_conda.models.environment import CondaEnvironment, Environment
from pdm_conda.models.setup import CondaSetupDistribution


class CondaSynchronizer(Synchronizer):
    def __init__(
        self,
        candidates: dict[str, Candidate],
        environment: Environment,
        clean: bool = False,
        dry_run: bool = False,
        retry_times: int = 1,
        install_self: bool = False,
        no_editable: bool | Collection[str] = False,
        use_install_cache: bool = False,
        reinstall: bool = False,
        only_keep: bool = False,
    ) -> None:
        super().__init__(
            candidates,
            environment,
            clean,
            dry_run,
            retry_times,
            install_self,
            no_editable,
            use_install_cache,
            reinstall,
            only_keep,
        )
        self.environment = cast(CondaEnvironment, environment)
        self.parallel = bool(self.parallel)  # type: ignore

    def compare_with_working_set(self) -> tuple[list[str], list[str], list[str]]:
        to_add, to_update, to_remove = super().compare_with_working_set()

        # deactivate parallel execution if uninstall
        self.parallel = self.environment.project.config["install.parallel"]
        if to_remove or to_update:
            if to_remove:
                to_remove = [p for p in to_remove if p not in self.environment.python_requirements]

            if self.parallel and any(True for d in to_remove + to_update if isinstance(d, CondaSetupDistribution)):
                self.environment.project.core.ui.echo("Deactivating parallel uninstall.")
                self.parallel = False

        return to_add, to_update, to_remove



from typing import Any, Iterable, Mapping, cast

from pdm._types import Source
from pdm.models.repositories import BaseRepository, LockedRepository, PyPIRepository
from pdm.models.requirements import Requirement
from pdm.models.specifiers import PySpecSet
from unearth import Link

from pdm_conda.models.candidates import Candidate, CondaCandidate
from pdm_conda.models.environment import CondaEnvironment, Environment
from pdm_conda.models.requirements import (
    CondaRequirement,
    NamedRequirement,
    parse_requirement,
)
from pdm_conda.plugin import conda_search


class CondaRepository(BaseRepository):
    def __init__(self, sources: list[Source], environment: Environment, ignore_compatibility: bool = True) -> None:
        super().__init__(sources, environment, ignore_compatibility)
        self.environment = cast(CondaEnvironment, environment)

    def get_dependencies(self, candidate: Candidate) -> tuple[list[Requirement], PySpecSet, str]:
        if isinstance(candidate, CondaCandidate):
            dependencies = list(candidate.dependencies)
            if candidate.constrains:
                for i, dep in enumerate(dependencies):
                    if (constrain := candidate.constrains.get(dep.identify(), None)) is not None:
                        dependencies[i] = constrain
            return (
                dependencies,
                PySpecSet(candidate.requires_python),
                candidate.summary,
            )
        requirements, requires_python, summary = super().get_dependencies(candidate)
        conda_conf = self.environment.project.conda_config
        if conda_conf.as_default_manager:
            for i, req in enumerate(requirements):
                if (
                    req.name not in conda_conf.excluded
                    and isinstance(req, NamedRequirement)
                    and not isinstance(req, CondaRequirement)
                ):
                    req.marker = None
                    req.name = req.conda_name
                    requirements[i] = parse_requirement(f"conda:{req.as_line()}")
        return requirements, requires_python, summary

    def get_hashes(self, candidate: Candidate) -> dict[Link, str] | None:
        if isinstance(candidate, CondaCandidate):
            return None
        return super().get_hashes(candidate)


class PyPICondaRepository(PyPIRepository, CondaRepository):
    def _find_candidates(self, requirement: Requirement) -> Iterable[Candidate]:
        if isinstance(requirement, CondaRequirement):
            candidates = conda_search(requirement, self.environment.project)
        else:
            candidates = super()._find_candidates(requirement)
        if (req := self.environment.python_requirements.get(requirement.conda_name, None)) is not None:
            candidates = [c for c in candidates if req.specifier.contains(c.version)]
        return candidates


class LockedCondaRepository(LockedRepository, CondaRepository):
    def _read_lockfile(self, lockfile: Mapping[str, Any]) -> None:
        packages = lockfile.get("package", [])
        conda_packages = [p for p in packages if p.get("conda_managed", False)]
        packages = [p for p in packages if not p.get("conda_managed", False)]
        super()._read_lockfile({"package": packages})

        for package in conda_packages:
            can = CondaCandidate.from_lock_package(package)
            can_id = self._identify_candidate(can)
            self.packages[can_id] = can
            self.candidate_info[can_id] = (
                [d.as_line(as_conda=True, with_build_string=True) for d in can.dependencies],
                package.get("requires_python", ""),
                "",
            )

    def _identify_candidate(self, candidate: Candidate) -> tuple:
        if isinstance(candidate, CondaCandidate):
            return candidate.identify(), candidate.version, None, False
        return super()._identify_candidate(candidate)



from pathlib import Path
from typing import cast

from pdm.core import Core
from pdm.exceptions import ProjectError
from pdm.models.environment import Environment
from pdm.models.repositories import LockedRepository
from pdm.models.specifiers import PySpecSet
from pdm.project import Project
from pdm.utils import get_venv_like_prefix
from tomlkit.items import Array

from pdm_conda.models.config import PluginConfig
from pdm_conda.models.requirements import (
    CondaRequirement,
    NamedRequirement,
    Requirement,
    parse_requirement,
)


class CondaProject(Project):
    def __init__(
        self,
        core: Core,
        root_path: str | Path | None,
        is_global: bool = False,
        global_config: str | Path | None = None,
    ) -> None:
        from pdm_conda.installers.manager import CondaInstallManager
        from pdm_conda.installers.synchronizers import CondaSynchronizer
        from pdm_conda.models.environment import CondaEnvironment
        from pdm_conda.models.repositories import (
            LockedCondaRepository,
            PyPICondaRepository,
        )
        from pdm_conda.resolver import CondaResolver

        super().__init__(core, root_path, is_global, global_config)
        self.core.repository_class = PyPICondaRepository
        self.core.install_manager_class = CondaInstallManager
        self.core.synchronizer_class = CondaSynchronizer
        self.core.resolver_class = CondaResolver
        self.locked_repository_class = LockedCondaRepository
        self.environment_class = CondaEnvironment
        self.virtual_packages: set[str] = set()
        self._conda_mapping: dict[str, str] = dict()
        self._pypi_mapping: dict[str, str] = dict()
        self.conda_config = PluginConfig.load_config(self)

    def get_conda_pyproject_dependencies(self, group: str, dev: bool = False) -> list[str]:
        """
        Get the conda dependencies array in the pyproject.toml
        """
        settings = self.pyproject.settings.setdefault("conda", dict())
        if group == "default":
            deps = settings.setdefault("dependencies", [])
        else:
            name = "optional" if not dev else "dev"
            deps = settings.setdefault(f"{name}-dependencies", dict()).setdefault(group, [])
        return deps

    def get_dependencies(self, group: str | None = None) -> dict[str, Requirement]:
        result = super().get_dependencies(group)

        config = self.conda_config
        group = group or "default"

        dev = group not in config.optional_dependencies
        if group in config.optional_dependencies and group in config.dev_dependencies:
            self.core.ui.echo(
                f"The {group} group exists in both [optional-dependencies] "
                "and [dev-dependencies], the former is taken.",
                err=True,
                style="warning",
            )
        deps = self.get_conda_pyproject_dependencies(group, dev)

        for line in deps:
            req = parse_requirement(f"conda:{line}")
            # search for package with extras to remove it
            pypi_req = next((v for v in result.values() if v.conda_name == req.conda_name), None)
            if pypi_req is not None:
                result.pop(pypi_req.identify())
                if not req.specifier:
                    req.specifier = pypi_req.specifier
            result[req.identify()] = req

        if self.conda_config.as_default_manager:
            for k in list(result):
                req = result[k]
                if (
                    req.name not in self.conda_config.excluded
                    and isinstance(req, NamedRequirement)
                    and not isinstance(req, CondaRequirement)
                ):
                    req.name = req.conda_name
                    result[k] = parse_requirement(f"conda:{req.as_line()}")

        return result

    @property
    def locked_repository(self) -> LockedRepository:
        try:
            lockfile = self.lockfile._data.unwrap()
        except ProjectError:
            lockfile = {}

        return self.locked_repository_class(lockfile, self.sources, self.environment)

    def add_dependencies(
        self,
        requirements: dict[str, Requirement],
        to_group: str = "default",
        dev: bool = False,
        show_message: bool = True,
    ) -> None:
        conda_requirements = {n: r for n, r in requirements.items() if isinstance(r, CondaRequirement)}
        requirements = {n: r for n, r in requirements.items() if n not in conda_requirements} | {
            n: r.as_named_requirement() for n, r in conda_requirements.items() if r.is_python_package
        }
        if self.conda_config.as_default_manager:
            conda_requirements = {n: r for n, r in conda_requirements.items() if not r.is_python_package}
        if conda_requirements:
            deps = self.get_conda_pyproject_dependencies(to_group, dev)
            cast(Array, deps).multiline(True)
            for _, dep in conda_requirements.items():
                matched_index = next((i for i, r in enumerate(deps) if dep.matches(f"conda:{r}")), None)
                req = dep.as_line(with_channel=True)
                if matched_index is None:
                    deps.append(req)
                else:
                    deps[matched_index] = req
            self.conda_config.reload()

        super().add_dependencies(requirements, to_group, dev, show_message)

    def get_environment(self) -> Environment:
        if not self.config["python.use_venv"]:
            raise ProjectError("python.use_venv is required to use Conda.")
        if get_venv_like_prefix(self.python.executable) is None:
            raise ProjectError("Conda environment not detected.")

        env = self.environment_class(self)
        if self.is_global:
            # Rewrite global project's python requires to be
            # compatible with the exact version
            env.python_requires = PySpecSet(f"=={self.python.version}")

        return env



import json
import subprocess
from functools import lru_cache
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import cast
from urllib.parse import urlparse

from pdm.exceptions import RequirementError
from pdm.models.setup import Setup
from pdm.project import Project

from pdm_conda.models.candidates import CondaCandidate
from pdm_conda.models.requirements import (
    CondaRequirement,
    Requirement,
    parse_conda_version,
    parse_requirement,
)
from pdm_conda.models.setup import CondaSetupDistribution
from pdm_conda.project import CondaProject
from pdm_conda.utils import normalize_name


def run_conda(cmd, **environment) -> dict:
    """
    Creates temporary environment file and run conda command
    :param cmd: conda command
    :param environment: environment data
    :return: conda command response
    """
    with NamedTemporaryFile(mode="w+", suffix=".yml") as f:
        if environment:
            for name, options in environment.items():
                if options:
                    f.write(f"{name}:\n")
                    for v in options:
                        f.write(f"  - {v}\n")
            f.seek(0)
            cmd = cmd + ["-f", f.name]
        process = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8")
        if "--json" in cmd:
            try:
                response = json.loads(process.stdout)
            except:
                response = {}
        else:
            response = {"message": process.stdout}
        try:
            process.check_returncode()
        except subprocess.CalledProcessError:
            msg = "Error locking dependencies\n"
            if isinstance(response, dict):
                if response.get("success", False):
                    pass
                if err := response.get("solver_problems", []):
                    msg += "\n".join(err)
                elif err := response.get("message", []):
                    msg += err
            else:
                msg += process.stderr
            msg += str(environment)
            raise RequirementError(msg)
    return response


@lru_cache(maxsize=None)
def _conda_search(
    requirement: str,
    project: CondaProject,
    channels: tuple[str],
) -> list[CondaCandidate]:
    """
    Search conda candidates for a requirement
    :param requirement: requirement
    :param project: PDM project
    :param channels: requirement channels
    :return: list of conda candidates
    """
    command = project.conda_config.command("search")
    if not project.virtual_packages:
        project.virtual_packages = conda_virtual_packages(project)

    command.append(requirement)
    for c in channels:
        command.extend(["-c", c])
    command.append("--json")
    result = run_conda(command)
    candidates = []
    # sort values per build number (greater first)
    packages = result.get("result", dict()).get("pkgs", [])
    _packages: dict[tuple, list] = dict()
    for p in packages:
        name = p["name"]
        version = p["version"]
        k = (name, version)
        _packages.setdefault(k, []).append(p)
    packages = []
    for ps in _packages.values():
        packages.extend(sorted(reversed(ps), key=lambda p: p["build_number"], reverse=True))

    for p in packages:
        dependencies = p.get("depends", [])
        valid_candidate = True
        for d in dependencies:
            if d.startswith("__"):
                if d not in project.virtual_packages:
                    valid_candidate = False
                    break
        if valid_candidate:
            package_channel = urlparse(p["channel"]).path
            for c in channels:
                if c in package_channel:
                    p["channel"] = c
                    break
            if "defaults" in channels and p["channel"].startswith("http"):
                p["channel"] = "defaults"

            candidates.append(CondaCandidate.from_conda_package(p))

    return candidates


def conda_search(
    requirement: CondaRequirement | str,
    project: CondaProject,
    channel: str | None = None,
) -> list[CondaCandidate]:
    """
    Search conda candidates for a requirement
    :param requirement: requirement
    :param project: PDM project
    :param channel: requirement channel
    :return: list of conda candidates
    """
    if isinstance(requirement, CondaRequirement):
        channel = channel or requirement.channel
        requirement = requirement.as_line(with_build_string=True).replace(" ", "=").replace("~", "")
    if "::" in requirement:
        channel, requirement = requirement.split("::", maxsplit=1)
    channels = [channel] if channel else (project.conda_config.channels or ["defaults"])
    return _conda_search(requirement, project, tuple(channels))


def update_requirements(requirements: list[Requirement], conda_packages: dict[str, CondaCandidate]):
    """
    Update requirements list with conda_packages
    :param requirements: requirements list
    :param conda_packages: conda packages
    """
    repeated_packages: dict[str, int] = dict()
    for i, requirement in enumerate(requirements):
        if (name := requirement.conda_name) in conda_packages and not isinstance(requirement, CondaRequirement):
            requirement.name = requirement.conda_name
            requirements[i] = parse_requirement(f"conda:{requirement.as_line()}")
        repeated_packages[name] = repeated_packages.get(name, 0) + 1
    to_remove = []
    for r in requirements:
        if repeated_packages.get(r.name, 1) > 1:
            to_remove.append(r)
            repeated_packages.pop(r.name)
    for r in to_remove:
        requirements.remove(r)


def lock_conda_dependencies(project: Project, requirements: list[Requirement], **kwargs):
    """
    Overwrite requirements with conda versions if needed in lock
    :param project: PDM project
    :param requirements: requirements list
    """
    project = cast(CondaProject, project)
    config = project.conda_config
    if not config.is_initialized:
        return

    _requirements = [r for r in requirements if isinstance(r, CondaRequirement)]
    if 0 < len(_requirements) < len(requirements):
        python_req = f"python=={project.python.version}"
        working_set = project.environment.get_working_set()
        if "python" in working_set:
            python_req = working_set["python"].as_line()
        _requirements.insert(0, parse_requirement(f"conda:{python_req}"))
        with TemporaryDirectory() as d:
            prefix = f"{d}/env"
            run_conda(config.command("create") + ["--prefix", prefix, "--json"])
            project.core.ui.echo(f"Created temporary environment at {prefix}")
            try:
                conda_packages = conda_lock(project, _requirements, prefix)
            finally:
                run_conda(config.command("remove") + ["--prefix", prefix, "--json"])
                project.core.ui.echo(f"Removed temporary environment at {prefix}")
    else:
        conda_packages = {}
        for r in _requirements:
            candidate = conda_search(r, project)[0]
            conda_packages[candidate.name] = candidate

    update_requirements(requirements, conda_packages)


def conda_lock(
    project: CondaProject,
    requirements: list[CondaRequirement],
    prefix: str,
) -> dict[str, CondaCandidate]:
    """
    Resolve conda marked requirements
    :param project: PDM project
    :param requirements: list of requirements
    :param prefix: environment prefix
    :return: resolved packages
    """
    packages: dict[str, CondaCandidate] = dict()
    config = project.conda_config
    core = project.core
    if not requirements:
        return packages

    core.ui.echo("Using conda to get: " + " ".join([r.as_line() for r in requirements if r.name != "python"]))
    _requirements = [r.as_line(with_build_string=True, with_channel=True).replace(" ", "=") for r in requirements]
    response = run_conda(
        config.command() + ["--force-reinstall", "--json", "--dry-run", "--prefix", prefix],
        channels=config.channels or ["defaults"],
        dependencies=_requirements,
    )

    if "actions" in response:
        for package in response["actions"]["LINK"]:
            package = CondaCandidate.from_conda_package(package)
            packages[package.name] = package
    else:
        if (msg := response.get("message", None)) is not None:
            core.ui.echo(msg)

    return packages


def _conda_install(
    project: CondaProject,
    command: list[str],
    packages: str | list[str] | None = None,
    verbose: bool = False,
):
    if isinstance(packages, str):
        packages = [packages]
    kwargs = dict()
    if packages:
        kwargs["dependencies"] = packages
    response = run_conda(command + ["--json"], **kwargs)
    if verbose:
        project.core.ui.echo(response)


def conda_install(
    project: CondaProject,
    packages: str | list[str],
    verbose: bool = False,
    dry_run: bool = False,
    no_deps: bool = False,
):
    """
    Install resolved packages using conda
    :param project: PDM project
    :param packages: resolved packages
    :param verbose: show conda response if true
    :param dry_run: don't install if dry run
    :param no_deps: don't install dependencies if true
    """
    config = project.conda_config
    command = config.command()
    if no_deps:
        command.append("--no-deps")
    if dry_run:
        command.append("--dry-run")
    if config.installation_method == "copy":
        _copy = "copy"
        if config.runner == "micromamba":
            _copy = f"always-{_copy}"
        command.append(f"--{_copy}")

    _conda_install(project, command, packages, verbose)


def conda_uninstall(
    project: CondaProject,
    packages: str | list[str],
    verbose: bool = False,
    dry_run: bool = False,
    no_deps: bool = False,
):
    """
    Uninstall resolved packages using conda
    :param project: PDM project
    :param packages: resolved packages
    :param verbose: show conda response if true
    :param dry_run: don't uninstall if dry run
    :param no_deps: don't uninstall dependencies if true
    """
    config = project.conda_config
    command = config.command("remove")
    if no_deps:
        _prune = "no-prune"
        if config.runner == "conda":
            _prune = "force-remove"
        command.append(f"--{_prune}")
    if dry_run:
        command.append("--dry-run")
    if isinstance(packages, str):
        packages = [packages]
    command.extend(packages)
    command.append("--json")

    _conda_install(project, command, verbose=verbose)


def conda_virtual_packages(project: CondaProject) -> set[str]:
    """
    Get conda virtual packages
    :param project: PDM project
    :return: set of virtual packages
    """
    config = project.conda_config
    virtual_packages = []
    if config.is_initialized:
        info = run_conda(config.command("info") + ["--json"])
        virtual_packages = [p.split("=")[0] for p in info["virtual packages"]]
    return set(virtual_packages)


def conda_list(project: CondaProject) -> dict[str, CondaSetupDistribution]:
    """
    List conda installed packages
    :param project: PDM project
    :return: packages distribution
    """
    config = project.conda_config
    distributions = dict()
    if config.is_initialized:
        packages = run_conda(config.command("list") + ["--json"])
        for package in packages:
            name, version = package["name"], package["version"]
            distributions[normalize_name(name)] = CondaSetupDistribution(
                Setup(
                    name=name,
                    summary="",
                    version=parse_conda_version(version),
                ),
                package=package,
            )

    return distributions



from importlib.metadata import Distribution
from typing import cast

from installer.exceptions import InstallerError
from pdm.exceptions import RequirementError, UninstallError
from pdm.installers import InstallManager

from pdm_conda.models.candidates import Candidate, CondaCandidate
from pdm_conda.models.environment import CondaEnvironment, Environment
from pdm_conda.models.setup import CondaSetupDistribution
from pdm_conda.plugin import conda_install, conda_uninstall


class CondaInstallManager(InstallManager):
    def __init__(self, environment: Environment, *, use_install_cache: bool = False) -> None:
        super().__init__(environment, use_install_cache=use_install_cache)
        self.environment = cast(CondaEnvironment, environment)

    def install(self, candidate: Candidate) -> None:
        """
        Install candidate, use conda if conda package else default installer
        :param candidate: candidate to install
        """
        if isinstance(candidate, CondaCandidate):
            try:
                conda_install(self.environment.project, candidate.link.url, no_deps=True)
            except (RequirementError, ValueError) as e:
                raise InstallerError(e) from e
        else:
            super().install(candidate)

    def uninstall(self, dist: Distribution) -> None:
        """
        Uninstall distribution, use conda if conda package else default uninstaller
        :param dist: distribution to uninstall
        """
        if isinstance(dist, CondaSetupDistribution):
            try:
                conda_uninstall(self.environment.project, dist.name, no_deps=True)
            except RequirementError as e:
                raise UninstallError(e) from e
        else:
            super().uninstall(dist)



import argparse
from typing import cast

from pdm.cli.commands.add import Command as BaseCommand
from pdm.cli.options import ArgumentGroup

from pdm_conda.models.requirements import CondaRequirement, parse_requirement
from pdm_conda.project import CondaProject, Project

conda_group = ArgumentGroup("Conda Arguments")
conda_group.add_argument(
    "--conda",
    dest="conda_packages",
    action="append",
    help="Specify Conda packages",
    default=[],
)

conda_group.add_argument(
    "-c",
    "--channel",
    dest="conda_channel",
    type=str,
    help="Specify Conda channel",
    default="",
)

conda_group.add_argument(
    "-r",
    "--runner",
    dest="conda_runner",
    type=str,
    help="Specify Conda runner executable",
    default="",
)


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "add"
    arguments = BaseCommand.arguments + [conda_group]

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        project = cast(CondaProject, project)
        config = project.conda_config
        if config.as_default_manager:
            options.conda_packages.extend(options.packages)
            options.packages = []

        if conda_packages := options.conda_packages:
            channel = options.conda_channel

            existing_channels = config.channels
            if options.conda_runner:
                config.runner = options.conda_runner
            if channel:
                existing_channels.append(channel)
                config.channels = existing_channels

            for package in conda_packages:
                package_channel = None
                if "::" in package:
                    package_channel, package = package.split("conda:", maxsplit=1)[-1].split("::", maxsplit=1)

                _p = parse_requirement(package)
                if isinstance(_p, CondaRequirement):
                    _p = _p.as_named_requirement()

                # if not named we can't use Conda
                if _p.is_named and _p.name not in config.excluded:
                    if package.startswith("conda:"):
                        package = package[len("conda:") :]
                    if not package_channel and channel:
                        package_channel = channel

                    if package_channel:
                        package = f"{package_channel}::{package}"
                        if package_channel not in existing_channels:
                            project.core.ui.echo(f"Detected Conda channel {package_channel}, adding it to pyproject")
                            existing_channels.append(package_channel)
                            config.channels = existing_channels

                    package = f"conda:{package}"

                options.packages.append(package)

        super().handle(project, options)



import dataclasses
import re
from typing import Any

from packaging.specifiers import SpecifierSet
from packaging.version import Version
from pdm.models.requirements import NamedRequirement, Requirement, T
from pdm.models.requirements import parse_requirement as _parse_requirement
from pdm.models.requirements import strip_extras

from pdm_conda.mapping import conda_to_pypi, pypi_to_conda
from pdm_conda.utils import normalize_name

_conda_meta_req_re = re.compile(r"conda:([\w\-_]+::)?(.+)$")
_specifier_re = re.compile(r"((<|>|=|~=|!=|==)+)(.+)")
_conda_specifier_star_re = re.compile(r"([\w.]+)\*")
_conda_version_letter_re = re.compile(r"(\d|\.)([a-z]+)(\d?)")

_patched = False


def correct_specifier_star(match):
    res = match.group(1)
    if res.endswith("."):
        res += "0"
    return res


def parse_conda_version(version):
    def correct_conda_version(match):
        digit, letter_specifier, follow_digit = match.groups()
        if letter_specifier in ("a", "b", "rc", "dev", "post") and follow_digit:
            letter_specifier += follow_digit
        else:
            letter_specifier = "".join(str(ord(letter)) for letter in letter_specifier)
            if digit != ".":
                letter_specifier = f".{letter_specifier}"
        return f"{digit}{letter_specifier}"

    return _conda_version_letter_re.sub(correct_conda_version, version)


@dataclasses.dataclass(eq=False)
class CondaRequirement(NamedRequirement):
    channel: str | None = None
    _is_python_package: bool = dataclasses.field(default=True, repr=False, init=False)
    version_mapping: dict = dataclasses.field(default_factory=dict, repr=False)
    mapping_excluded: bool = dataclasses.field(default=False, repr=False)
    build_string: str | None = None

    @property
    def conda_name(self) -> str | None:
        return self.name

    @property
    def is_python_package(self):
        return self._is_python_package

    @is_python_package.setter
    def is_python_package(self, value):
        self._is_python_package = value
        self.mapping_excluded = not value

    @classmethod
    def create(cls: type[T], **kwargs: Any) -> T:
        kwargs.pop("conda_managed", None)
        if build_string := kwargs.get("build_string", None):
            kwargs["build_string"] = build_string.strip()

        return super().create(**kwargs)

    def as_line(self, as_conda: bool = False, with_channel=False, with_build_string=False) -> str:
        channel = f"{self.channel}::" if with_channel and self.channel else ""
        if as_conda:
            channel = f"conda:{channel}"
        specifier = ""
        for s in self.specifier:
            if specifier:
                specifier += ","
            specifier += f"{s.operator}{self.version_mapping.get(s.version, s.version)}"
        build_string = f" {self.build_string}" if with_build_string and self.build_string and specifier else ""
        return f"{channel}{self.conda_name}{specifier}{build_string}"

    def _hash_key(self) -> tuple:
        return (
            self.key,
            frozenset(self.specifier),
        )

    def as_named_requirement(self) -> NamedRequirement:
        return NamedRequirement.create(name=conda_to_pypi(self.name), specifier=self.specifier)


def remove_operator(version):
    return _specifier_re.sub(r"\3", version)


def parse_requirement(line: str, editable: bool = False) -> Requirement:
    if (match := _conda_meta_req_re.match(line)) is not None:
        version_mapping = dict()
        channel, line = match.groups()
        if channel:
            channel = channel[:-2]

        build_string = None
        if len(_line := line.split(" ")) == 3 or (len(_line) == 2 and _specifier_re.search(_line[0])):
            line = " ".join(_line[:-1])
            build_string = _line[-1]

        name = line
        version = ""
        if match := _specifier_re.search(line):
            name, version = line[: match.start(1)], line[match.start(1) :]
        elif " " in line:
            name, version = line.split(" ", maxsplit=1)
        version_and = version.split(",")
        for i, conda_version in enumerate(version_and):
            version_or = conda_version.split("|")
            for j, conda_version_or in enumerate(version_or):
                if conda_version_or:
                    if conda_version_or == "*":
                        _version = ""
                    else:
                        if not _specifier_re.match(conda_version_or):
                            s = "="
                            if _conda_specifier_star_re.match(conda_version_or):
                                s = "~"
                            conda_version_or = f"{s}={conda_version_or}"
                        _version = parse_conda_version(
                            _conda_specifier_star_re.sub(correct_specifier_star, conda_version_or),
                        )
                        version_mapping[remove_operator(_version)] = remove_operator(conda_version_or)
                    version_or[j] = _version
            version_and[i] = max((v for v in version_or if v), key=lambda v: Version(remove_operator(v)), default="")
        version = ",".join(version_and)
        req = CondaRequirement.create(
            name=strip_extras(name.strip())[0],
            specifier=SpecifierSet(version),
            channel=channel,
            version_mapping=version_mapping,
            build_string=build_string,
        )
    else:
        req = _parse_requirement(line=line, editable=editable)
    return req


def conda_name(self) -> str | None:
    if not hasattr(self, "_conda_name"):
        name = self.name
        self._conda_name = pypi_to_conda(strip_extras(name)[0]) if name else None
    return self._conda_name


def key(self) -> str | None:
    return normalize_name(self.conda_name).lower() if self.conda_name else None


if not _patched:
    from pdm.cli import actions
    from pdm.models import requirements

    for m in [actions, requirements]:
        setattr(m, "parse_requirement", parse_requirement)

    setattr(Requirement, "conda_name", property(conda_name))
    setattr(Requirement, "key", property(key))
    _patched = True



from pdm.core import Core


def main(core: Core):
    from pdm.signals import pre_lock

    from pdm_conda.cli.commands.add import Command as AddCommand
    from pdm_conda.cli.commands.remove import Command as RemoveCommand
    from pdm_conda.models.config import CONFIGS
    from pdm_conda.plugin import lock_conda_dependencies
    from pdm_conda.project import CondaProject

    from pdm_conda import utils  # noqa
    from pdm_conda.cli import utils as cli_utils  # noqa

    core.project_class = CondaProject

    core.register_command(AddCommand)
    core.register_command(RemoveCommand)

    pre_lock.connect(lock_conda_dependencies)
    for name, config in CONFIGS:
        core.add_config(name, config)


__version__ = "0.5.3"
